#!/usr/bin/env python3

from equity import Equity

from ..config import bytom

# Bytom configuration
bytom = bytom()

# Hash Time Lock Contract (HTLC) Script
HTLC_SCRIPT = """
contract HTLC (
  secret_hash: Hash,
  recipient: PublicKey,
  sender: PublicKey,
  sequence: Integer
) locks valueAmount of valueAsset {
  clause claim(preimage: String, sig: Signature) {
    verify sha256(preimage) == secret_hash
    verify checkTxSig(recipient, sig)
    unlock valueAmount of valueAsset
  }
  clause refund(sig: Signature) {
    verify above(sequence)
    verify checkTxSig(sender, sig)
    unlock valueAmount of valueAsset
  }
}
"""


# Hash Time Lock Contract
class HTLC:
    """
    Bytom Hash Time Lock Contract (HTLC) class.

    :param network: bytom network, defaults to testnet.
    :type network: str
    :returns:  HTLC -- bytom HTLC instance.

    .. note::
        Bytom has only three networks, ``mainnet``, ``solonet`` and ``testnet``.
    """

    # Initialization
    def __init__(self, network="testnet"):
        # Bytom network.
        self.network = network
        # Initializing equity
        self.equity = None

    # Initialize new HTLC Contract script
    def init(self, secret_hash, recipient_public, sender_public, sequence=bytom["sequence"]):
        """
        Initialize bytom Hash Time Lock Contract (HTLC).

        :param secret_hash: secret sha-256 hash.
        :type secret_hash: hash
        :param recipient_public: bytom recipient public key.
        :type recipient_public: str
        :param sender_public: bytom sender public key.
        :type sender_public: int
        :param sequence: bytom sequence number of expiration block, defaults to bytom config sequence (15).
        :type sequence: int
        :returns: HTLC -- bytom Hash Time Lock Contract (HTLC) instance.

        >>> from shuttle.providers.bytom.htlc import HTLC
        >>> htlc = HTLC(network="testnet")
        >>> htlc.init(secret_hash, recipient_public_key, sender_public_key, 100)
        <shuttle.providers.bytom.htlc.HTLC object at 0x0409DAF0>
        """

        # HTLC agreements
        HTLC_ARGS = [
            secret_hash,  # secret_hash: Hash
            recipient_public,  # recipient: PublicKey
            sender_public,  # sender: PublicKey
            sequence,  # sequence: Integer
        ]
        # Compiling HTLC contract
        self.equity = Equity(bytom[self.network]["bytom"])\
            .compile_source(HTLC_SCRIPT, HTLC_ARGS)
        return self

    # Bytecode HTLC script
    def bytecode(self):
        """
        Get bytom htlc bytecode.

        :returns: str -- bytom Hash Time Lock Contract (HTLC) bytecode.

        >>> from shuttle.providers.bytom.htlc import HTLC
        >>> htlc = HTLC(network="testnet")
        >>> htlc.init(secret_hash, recipient_public_key, sender_public_key, 100)
        >>> htlc.bytecode()
        "01642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c0"
        """

        if self.equity and "program" not in self.equity:
            raise ValueError("htlc script is none, initialization htlc first")
        return self.equity["program"]

    # OP Code of HTLC script
    def opcode(self):
        """
        Get bytom htlc opcode.

        :returns: str -- bytom Hash Time Lock Contract (HTLC) opcode.

        >>> from shuttle.providers.bytom.htlc import HTLC
        >>> htlc = HTLC(network="testnet")
        >>> htlc.init(secret_hash, recipient_public_key, sender_public_key, 100)
        >>> htlc.opcode()
        "0x64 0x91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2 0xac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01 0x2b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff DEPTH 0x547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac FALSE CHECKPREDICATE"
        """

        if self.equity and "opcodes" not in self.equity:
            raise ValueError("htlc script is none, initialization htlc first")
        return self.equity["opcodes"]



#!/usr/bin/env python3

from base64 import b64encode, b64decode

import ed25519
import json
import binascii
import datetime

from .rpc import get_transaction, decode_tx_raw, submit_payment


def sign(private_key_str, message_str):
    signing_key = ed25519.SigningKey(bytes.fromhex(private_key_str))
    # signature = signing_key.sign(message_str.encode(), encoding='hex')
    signature = signing_key.sign(bytes.fromhex(message_str), encoding='hex')
    return signature.decode()


def verify(public_key_str, signature_str, message_str):
    result = False
    verifying_key = ed25519.VerifyingKey(public_key_str.encode(), encoding='hex')
    try:
        verifying_key.verify(signature_str.encode(), bytes.fromhex(message_str), encoding='hex')
        result = True
    except ed25519.BadSignatureError:
        result = False
    return result


def find_contract_utxo_id(tx_id, network):
    """
    Find smart contract UTXO id.

    :param tx_id: bytom transaction id or hash.
    :type tx_id: str
    :param network: bytom network.
    :type network: str
    :returns: str -- UTXO id.

    >>> from shuttle.providers.bytom.utils import find_contract_utxo_id
    >>> find_contract_utxo_id(bytom_transaction_id, "mainnet")
    "9059cd0d03e4d4fab70a415169a45be47583f7240115c36cf298d6f261c0a1ac"
    """

    utxo_id = None
    contract_transaction = get_transaction(tx_id=tx_id, network=network)
    contract_outputs = contract_transaction["outputs"]
    for contract_output in contract_outputs:
        if contract_output["address"] == "smart contract":
            utxo_id = contract_output["utxo_id"]
            break
    return utxo_id


def decode_transaction_raw(tx_raw):
    """
    Decode bytom transaction raw.

    :param tx_raw: bytom transaction raw.
    :type tx_raw: str
    :returns: dict -- decoded bytom transaction.

    >>> from shuttle.providers.bytom.utils import decode_transaction_raw
    >>> decode_transaction_raw(transaction_raw)
    {...}
    """

    tx_raw = str(tx_raw + "=" * (-len(tx_raw) % 4))
    try:
        # Decoding transaction raw.
        decoded_tx_raw = json.loads(b64decode(str(tx_raw).encode()).decode())
    except (binascii.Error, json.decoder.JSONDecodeError) as _error:
        raise ValueError("invalid bytom transaction raw")
    if "type" not in decoded_tx_raw and not str(decoded_tx_raw["type"]).startswith("bytom"):
        raise ValueError("invalid bytom transaction raw")
    return dict(
        fee=decoded_tx_raw["fee"],
        guid=decoded_tx_raw["guid"],
        type=decoded_tx_raw["type"],
        tx=decode_tx_raw(tx_raw=decoded_tx_raw["raw"]),
        unsigned=decoded_tx_raw["unsigned"],
        signatures=decoded_tx_raw["signatures"],
        network=decoded_tx_raw["network"]
    )


def submit_transaction_raw(tx_raw):
    """
    Submit transaction raw to Bytom blockchain.

    :param tx_raw: bytom transaction raw.
    :type tx_raw: str
    :returns: dict -- bytom transaction id, fee, type and date.

    >>> from shuttle.providers.bytom.utils import submit_transaction_raw
    >>> submit_transaction_raw(transaction_raw)
    {...}
    """

    tx_raw = str(tx_raw + "=" * (-len(tx_raw) % 4))
    try:
        # Decoding transaction raw.
        decoded_tx_raw = json.loads(b64decode(str(tx_raw).encode()).decode())
    except (binascii.Error, json.decoder.JSONDecodeError) as _error:
        raise ValueError("invalid bytom transaction raw")
    if "type" not in decoded_tx_raw and not str(decoded_tx_raw["type"]).startswith("bytom"):
        raise ValueError("invalid bytom transaction raw")
    submitted = submit_payment(
        guid=decoded_tx_raw["guid"],
        tx_raw=decoded_tx_raw["raw"],
        signatures=decoded_tx_raw["signatures"],
        network=decoded_tx_raw["network"]
    )
    return dict(
        fee=decoded_tx_raw["fee"],
        type=decoded_tx_raw["type"],
        tx_id=submitted["transaction_hash"],
        network=decoded_tx_raw["network"],
        date=str(datetime.datetime.utcnow())
    )


def spend_utxo_action(utxo):
    """
    Get spend UTXO action

    :param utxo: bytom butxo id.
    :type utxo: str
    :returns: dict -- bytom spend utxo action.

    >>> from shuttle.providers.bytom.utils import spend_utxo_action
    >>> spend_utxo_action(bytom_utxo_id)
    {...}
    """

    return dict(type=str("spend_utxo"), output_id=utxo)


def contract_arguments(amount, address):
    """
    Get contract arguments.

    :param amount: bytom amount.
    :type amount: int
    :param address: bytom address.
    :type address: str
    :returns: list -- bytom contract arguments.

    >>> from shuttle.providers.bytom.utils import contract_arguments
    >>> contract_arguments(bytom_amount, bytom_address)
    [...]
    """

    return [dict(type=str("integer"), value=amount),
            dict(type=str("address"), value=address), dict(type=str("data"), value=str())]


def spend_wallet_action(amount, asset):
    """
    Get spend wallet action.

    :param amount: bytom amount.
    :type amount: int
    :param asset: bytom asset.
    :type asset: str
    :returns: dict -- bytom spend wallet action.

    >>> from shuttle.providers.bytom.utils import spend_wallet_action
    >>> spend_wallet_action(bytom_amount, bytom_asset)
    {...}
    """

    return dict(amount=amount,
                asset=asset, type=str("spend_wallet"))


def spend_account_action(account, amount, asset):
    """
    Get spend account action.

    :param account: bytom account.
    :type account: str
    :param amount: bytom amount.
    :type amount: int
    :param asset: bytom asset.
    :type asset: str
    :returns: dict -- bytom spend account action.

    >>> from shuttle.providers.bytom.utils import spend_account_action
    >>> spend_account_action(bytom_account, bytom_amount, bytom_asset)
    {...}
    """

    return dict(account=account,
                amount=amount, asset=asset, type=str("spend_account"))


def control_program_action(amount, asset, control_program):
    """
    Get control program action.

    :param amount: bytom amount.
    :type amount: int
    :param asset: bytom asset.
    :type asset: str
    :param control_program: bytom control program.
    :type control_program: str
    :returns: dict -- bytom control program action.

    >>> from shuttle.providers.bytom.utils import control_program_action
    >>> control_program_action(bytom_amount, bytom_asset, bytom_control_program)
    {...}
    """

    return dict(amount=amount, asset=asset,
                control_program=control_program, type=str("control_program"))


def control_address_action(amount, asset, address):
    """
    Get control address action.

    :param amount: bytom amount.
    :type amount: int
    :param asset: bytom asset.
    :type asset: str
    :param address: bytom address.
    :type address: str
    :returns: dict -- bytom control address action.

    >>> from shuttle.providers.bytom.utils import control_address_action
    >>> control_address_action(bytom_amount, bytom_asset, bytom_address)
    {...}
    """

    return dict(amount=amount, asset=asset,
                address=address, type=str("control_address"))



#!/usr/bin/env python3

from setuptools import setup, find_packages

import shuttle

# README.md
with open("README.md", "r") as readme:
    long_description = readme.read()

# requirements.txt
with open("requirements.txt", "r") as _requirements:
    requirements = list(map(str.strip, _requirements.read().split("\n")))[:-1]

setup(
    name="pyshuttle",
    version=shuttle.__version__,
    description="Cross-chain atomic swaps between the networks of two cryptocurrencies.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=shuttle.__license__,
    author=shuttle.__author__,
    author_email=shuttle.__email__,
    url="https://github.com/meherett/shuttle",
    packages=find_packages(),
    keywords=["shuttle"],
    entry_points={
        "console_scripts": ["shuttle=shuttle.cli.__main__:main"]
    },
    python_requires=">=3.6,<4",
    install_requires=requirements,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
)
